<?php
/**
 * Cms_scheduler.php
 *
 * The scheduler to run startup routines and scheduled events such as PDF cleanup.
 * This relies on various hooks placed around the site. We piggy back on user requests
 * to run our scheduler events. This can not be used if precise timing for a event is needed.
 * This will not run at all if there are no user logged in. We could make a process/thread to stay
 * in memory all the time but preferred not to. Some ISPs might have a cleanup script to kill
 * processes that are run too long...
 *
 * All event functions are required to do an exit at the end. DO NOT USE osa_exit() here.
 * This will discard any buffered output so it is not suitable for most scripts. Only a page that
 * does not print any output to the broswer can call this library.
 *
 * @author $Author: dtong $
 * @version $Id: Cms_scheduler.php,v 1.3 2012/05/23 06:46:51 dtong Exp $
 * @copyright Copyright (c) 2012, Tiller Software Co., Ltd.
*/

class Cms_scheduler {
	// Making this an hourly event on purpose so we can add week, day, minute events later.
	const CONFIG_NAME_HOUR = 'cms_scheduler_run_time_hour';
	// This determines how often the events will get run.
	// Note, the login event will always run during each successful login.
	const TIME_HOUR = 3600;
	// Delete tmp files if they are older than these values. 14400 = 4 hours.
	const FILE_EXPIRE_TIME_PDF = 14400;
	const FILE_EXPIRE_TIME_UNITZIP = 14400;
	const FILE_EXPIRE_UPLOAD_TMP = 14400;
	const FILE_EXPIRE_TIME_PDF_HEADER_FOOTER_HTML = 1800;

	private static $event_functions = array('_event_tmpfile_cleanup', '_event_3p');
	private static $init_ready = FALSE;

	private $CI;

	function __construct() {
		$this->CI = &get_instance();
	}

	//This is a major system class so protect it from lazy coders...
	function __set($var, $value) {
		// You will most likely not see the message so write it to the log too.
		osa_log(__METHOD__, "You are not allowed to set undeclared class property '$var', bye bye.");
		echo __METHOD__ . " - You are not allowed to set undeclared class property '$var', bye bye.";
		exit;
	}

	// Event to fire up during an successful login. The login does a redirect to home and
	// therefore no html output and perfect to run this event.
	function login_event() {
		$this->_init();
		osa_log_debug(__METHOD__, 'Starting login event.');
		// Load the template cache library. This will create the template cache if there isn't one yet.
		// For a new system, the first login will create this cache
		// The library param can be anything, it is there to tell the lib to check for mod time
		// If the template files or related source files got updated then the cache would get updated too.
		$this->CI->load->library('cms/cms_template/cms_template_init', array('checkmodtime'=>TRUE));

		// DO NOT PUT the unit_cache_table code inside a lock in this class because unit_cache_table also
		// uses locking and one session can only hold one lock, this is mysql's way of preventing deadlocks.
		// This will initilize the unit_search_cache memory tables
		// The check_empty is a dummy array to tell the object to check for empty tables
		$this->CI->load->library('cms/memory_table/unit_cache_table', array('check_empty'=>TRUE));
		// Run the scheduled auto update if necessary
		$this->CI->unit_cache_table->auto_update();
		// Run the main event too.

		$this->event(FALSE);
		osa_log_debug(__METHOD__, 'End login event.');
		exit;
	}

	// This is the main scheduler event. Right now, we only have an hourly event.
	// But this function can be easily converted to accpet a time parameter (self::TIME_HOUR and self::CONFIG_NAME_HOUR)
	// The is run by the session check and login event for now.
	function event($do_exit=TRUE) {
		static $ran_once = FALSE;
		if ( $ran_once ) {
			return;
		}
		$config_time = $this->CI->config->item(self::CONFIG_NAME_HOUR);
		$cur_time = time();

		// Most requests will not go into this condition. This happens only once every hour.
		// See who is the unlucky one :).
		if ( $config_time === FALSE || ($cur_time - $config_time) >= self::TIME_HOUR) {
			$this->CI->load->model('cms/memory_table_model');
			if ( $this->CI->memory_table_model->lock(__METHOD__, 0) ) {

				// Let's get the DB value again. This does not happen often so overhead is kept to the minimum.
				if ( isset($this->CI->admin_loadconfig) ) {
					// Get the latest DB config values.
					$this->CI->config->set_item('cms_lang_ignore_all', TRUE);
					$this->CI->admin_loadconfig->load_config(config_data::SYSTEM_GENERAL, TRUE, FALSE);
					$this->CI->config->set_item('cms_lang_ignore_all', FALSE);
					if ( $config_time && $this->CI->admin_loadconfig->get_value(self::CONFIG_NAME_HOUR) != $config_time ) {
						// Last run time changed so another process already did the update so unlock and return.
						// This makes sure we never run the update twice at the same time. Because there is so much code between here and
						// where the config gets initialy read so reading it again after getting the lock ensures the value is the latest.
						$this->CI->memory_table_model->unlock(__METHOD__);
						$ran_once = TRUE;
						return;
					}
				}

				// Run the events
				$this->_init();
				osa_log_debug(__METHOD__, 'Starting scheduler event.');
				// Once we are sure we can go ahead with the scheduler event, update the last run value so
				// other processes will ignore the scheduler.
				osa_config_update2db(self::CONFIG_NAME_HOUR, $cur_time);
				try {
					foreach ( self::$event_functions as $func ) {
						$this->$func();
					}
				}
				catch (Exception $e) {
					// Something went wrong so reset the hour config and the next event run will do the update.
					osa_config_update2db(self::CONFIG_NAME_HOUR, FALSE);
					osa_log(__METHOD__, 'Got an exeption.', $e);
				}
				$this->CI->memory_table_model->unlock(__METHOD__);
				if ( $do_exit ) {
					osa_log_debug(__METHOD__, 'Ending scheduler event.');
					exit;
				}
			}
			// Do nothing if it failed to get the lock becuase it means another process is running the event.
		}
		$ran_once = TRUE;
		return;
	}

	private function _event_tmpfile_cleanup() {
		osa_log_debug(__METHOD__, 'Cleaning tmp files.');

		// PDF
		$folder = $this->CI->config->item('cms_pdf_tmp_dir');
		// Make sure we have a valid path
		if ( !empty($folder) && strlen($folder) > 6 ) {
			// This should not contain sub-directories so 4th param is true to not delete any sub-dirs.
			osa_remove_directory($folder, self::FILE_EXPIRE_TIME_PDF, TRUE, TRUE, '*.pdf');
		}

		// Zipdownload
		$folder = $this->CI->config->item('cms_resource_uploadtmpfolder');
		if ( !empty($folder) && strlen($folder) > 3 ) {
			$folder .= '/zipdownload';
			// This should not contain sub-directories so 4th param is true to not delete any sub-dirs.
			osa_remove_directory($folder, self::FILE_EXPIRE_TIME_UNITZIP, TRUE, TRUE, '*.zip');
		}

		// Upload tmp
		$folder = $this->CI->config->item('cms_resource_uploadtmpfolder');
		if ( !empty($folder) && strlen($folder) > 3 ) {
			$folder .= '/' . CMS_UPLOAD_FOLDR_NAME;
			osa_remove_directory($folder, self::FILE_EXPIRE_UPLOAD_TMP, TRUE, FALSE, '*.*');
		}

		// PDF header and footer files
		$folder = $this->CI->config->item('cms_header_footer_create_folder');
		if ( !empty($folder) && strlen($folder) > 3 ) {
			osa_remove_directory($folder, self::FILE_EXPIRE_TIME_PDF_HEADER_FOOTER_HTML, TRUE, TRUE, '*.html');
		}

		osa_log_debug(__METHOD__, 'Done cleaning tmp files.');
	}

	private function _event_3p() {
		osa_log_debug(__METHOD__, 'Running _event_3p().');
		if ( $obj = osa_3p_load_lib('CMS_3P_LIB_QUEUE', 'CMS_3P_FOLDER_COMMON') ) {
			$obj->runall();
		}
		osa_log_debug(__METHOD__, 'Done running _event_3p().');
	}

	// Always call this first in all event functions.
	// Once this function is called all output will be ignored and you won't see anything
	// from here on in the browser.
	private function _init() {
		if ( !self::$init_ready ) {
			// Run these 2 again to make sure.
			@ignore_user_abort();
			@set_time_limit(0);
			// If server is using compression then size has to be zero else it will wait until exit.
			// Size is always zero and all buffered output will be discarded.
			header('Content-Length: 0');
			header('Connection: close');
			// Without this line the browser won't terminate if server compression is on.
			header('Content-Encoding: none');
			@ob_end_clean();
			@flush();
		}
		self::$init_ready = TRUE;
	}
}